#!/usr/bin/env python
# vim: expandtab sw=4 sts=0:

"""
@copyright: Copyright (C) 2008-2016 Oracle.  All rights reserved.

@license:   This program is free software; you can redistribute it and/or
            modify it under the terms of the GNU General Public
            License as published by the Free Software Foundation, version 2.

            This program is distributed in the hope that it will be useful,
            but WITHOUT ANY WARRANTY; without even the implied warranty of
            MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
            General Public License for more details.

@summary:   This is the Oracle OCFS2 Plugin for Oracle Storage Connect.
"""

__docformat__ = 'epytext en'

import re
import os
import stat
import time
import subprocess
import errno
import struct
import platform
import uuid
import fnmatch
import xml.dom.minidom
import glob
from fcntl import ioctl
from gettext import gettext
from cStringIO import StringIO

from OSCPlugin import *
from OSCPluginUtils import (getPhysicalDiskType, deviceTypeNVMe,
                            getActiveDevMapper, getActiveDevPP, deviceTypeMD)

_ = gettext

__all__ = ['OCFS2Plugin']

class OCFS2Plugin(IFileSystemPlugin):
    """This is the Oracle supplied Plugin for the OCFS2 clustered file system"""

    #
    # Private class variables
    #
    _ocfs2_min_vers_        = "1.6.0"
    _osc_min_vers__         = ["1", "2", "7"]
    __plugin_release__      = "47.5.1"

    #
    # Required class variables as per the Storage Connect Plugin spec
    #
    plugin_name             = "Oracle OCFS2 File system"
    vendor_name             = "Oracle"
    plugin_version          = "0.1.0-" + __plugin_release__
    plugin_desc             = "Oracle OCFS2 File system Storage Connect Plugin"
    required_api_vers       = _osc_min_vers__
    filesys_type            = IFileSystemPlugin.LocalFileSystem
    filesys_name            = "ocfs2"
    plugin_ability          = {"snapshot":              ABILITY_TYPES.ONLINE,
                               "custom_snap_name":      ABILITY_TYPES.YES,
                               "snap_is_sync":          ABILITY_TYPES.YES,
                               "clone":                 ABILITY_TYPES.ONLINE,
                               "custom_clone_name":     ABILITY_TYPES.YES,
                               "clone_is_sync":         ABILITY_TYPES.YES,
                               "resize":                ABILITY_TYPES.ONLINE,
                               "resize_is_sync":        ABILITY_TYPES.YES,
                               "splitclone":            ABILITY_TYPES.UNSUPPORTED,
                               "splitclone_is_sync":    ABILITY_TYPES.NO,
                               "splitclone_while_open": ABILITY_TYPES.NO,
                               "require_storage_name":  ABILITY_TYPES.NO,
                               "snapclone":             ABILITY_TYPES.ONLINE,
                               "snapclone_is_sync":     ABILITY_TYPES.YES,
                               "backing_device_type":   BACKING_DEVICE_TYPES.DEVICE_SINGLE,
                               "access_control":        ABILITY_TYPES.NO}


    __cluster_required__    = True

    # Force 4K blocks so that size calculation is easy
    FS_BLOCKSIZE = 4096

    # Support up to 32 nodes
    FS_NODESLOTS = 32

    @staticmethod
    def validate(ss_record):
        ss_record = _fillSSRecord(**ss_record)
        try:
            _validateSSRecord(ss_record)
        except (MissingKeyEx, InvalidValueEx), e:
            error = _('Invalid SSrecord: %s') % str(e)
            IPlugin.logger.error(error)
            raise InvalidStorageArrayEx(error)

    @staticmethod
    def getCapabilities(ss_record):
        return OCFS2Plugin.plugin_ability;

    @staticmethod
    def getStorageNames(ss_record):
        return []

    @staticmethod
    def getAccessGroups(ss_record, fs_record=None):
        return [{'grp_name': 'all', 'grp_entries': ['all']}]

    @staticmethod
    def getQoSList(ss_record, fs_record):
        return [
                {'priority': 0, 'value': 'vmstore', 'description': 'Virtual machine images'},
                {'priority': 1, 'value': 'datafiles', 'description': 'Large data files'},
                {'priority': 2, 'value': 'mail', 'description': 'Lots of small files'},
               ]

    @staticmethod
    def getStorageServerInfo(ss_record):
        ret_record = _fillSSInfo(ss_record)
        return ret_record

    @staticmethod
    def getFileSystemInfo(ss_record, fs_record):
        ret_record = _fillFSInfo(ss_record, fs_record)
        return ret_record

    @staticmethod
    def getStatus(ss_record, fs_record=None):
        if fs_record:
            status = _getFSStatus(ss_record, fs_record)
        else:
            status = _getModuleStatus()
        return status

    @staticmethod
    def listFileSystems(ss_record, include_unpresented=False):
        #
        # Local filesystems have no concept of presented vs unpresented,
        # so all filesystems are returned.
        #
        _validateSSRecord(ss_record)
        fs_hash = {}
        dev_info = {}
        mounts = _getMounts()
        devices = _getOCFS2Devices()

        #
        # Get mounted devices first, so we can honor the particular
        # device mounted.
        #
        nvme_list = glob.glob(os.path.join(IPlugin.dev_path_by_id, 'nvme-*'))
        md_list = glob.glob(os.path.join(IPlugin.dev_path_by_id, 'md-uuid-*'))
        for (devname, mlist) in mounts.iteritems():
            if not dev_info.has_key(devname):
                try:
                    dev_info[devname] = _getOneOCFS2Device(devname)
                except:
                    continue
            info = dev_info[devname]
            if fs_hash.has_key(info['uuid']):
                continue
            fs_record = _deviceToFSRecord(ss_record, info)
            # mount returns device names such as /dev/nvme0 and /dev/md0 
            if len(nvme_list) > 0 and deviceTypeNVMe(fs_record['backing_device']):
                nvme_dev = fs_record['backing_device'].replace('/dev/', '')
                for dev in nvme_list:
                    if nvme_dev in os.readlink(dev):
                        fs_record['backing_device'] = dev
                        break
            if len(md_list) > 0 and deviceTypeMD(fs_record['backing_device']):
                md_dev = fs_record['backing_device'].replace('/dev/', '')
                for dev in md_list:
                    if md_dev in os.readlink(dev):
                        fs_record['backing_device'] = dev
                        break

            fs_record['status'] = 'mounted'
            fs_record['state'] = FS_STATES.MOUNTED
            fs_record['mount_options'] = mlist[0]['options'].split(',')
            fs_hash[info['uuid']] = fs_record

        # Now find unmounted filesystems
        for devname in devices:
            if dev_info.has_key(devname):
                continue
            try:
                dev_info[devname] = _getOneOCFS2Device(devname)
            except:
                continue
            info = dev_info[devname]
            if fs_hash.has_key(info['uuid']):
                continue
            fs_record = _deviceToFSRecord(ss_record, dev_info[devname])
            fs_record['status'] = 'not mounted'
            fs_record['state'] = FS_STATES.UNMOUNTED
            fs_hash[info['uuid']] = fs_record
        return fs_hash.values()

    @staticmethod
    def listMountPoints(ss_record, fs_record=None):
        _validateSSRecord(ss_record)
        if fs_record:
            _validateFSRecord(ss_record, fs_record)
        mounts = _getMounts()
        mount_records = []

        for (devname, mlist) in mounts.iteritems():
            info = _getOneOCFS2Device(devname)
            if fs_record and info['uuid'] != fs_record['uuid']:
                continue
            for m in mlist:
                mr = _deviceAndMountToMountRecord(info, m)
                mount_records.append(mr)
        return mount_records

    @staticmethod
    def mount(ss_record, fs_record, mountpoint, share_path=None,
              create_mount_point=False, mount_options=[]):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        if not mountpoint:
            error = _('Cannot mount filesystem "%s" without a mountpoint') % \
                fs_record['backing_device']
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)
        if share_path is not None:
            error = _('Share paths are not supported')
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)
        if create_mount_point and not os.path.exists(mountpoint):
            os.makedirs(mountpoint)
        mounts = OCFS2Plugin.listMountPoints(ss_record, fs_record)
        for mount in mounts:
            if os.path.samefile(mount['mount_point'], mountpoint):
                return

        cmd = ['mount']
        if len(mount_options):
            option_string = ','.join(mount_options)
            if len(option_string):
                cmd.extend(['-o', option_string])
        cmd.extend([fs_record['backing_device'], mountpoint])
        mnt_p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
        (p_out, p_err) = mnt_p.communicate()
        if mnt_p.returncode != 0:
            error = _('Unable to mount file system "%s": %s') % (fs_record['backing_device'],
                                                                 p_err)
            IPlugin.logger.error(error)
            raise OperationFailedEx(error)
        fs_record['mount_options'] = mount_options
        return fs_record

    @staticmethod
    def unmount(ss_record, fs_record, mountpoint, destroy_mount_point=False):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        mounts = OCFS2Plugin.listMountPoints(ss_record, fs_record)
        for mount in mounts:
            if os.path.samefile(mount['mount_point'], mountpoint):
                break
        else:
            IPlugin.logger.info('no mount record for mount point: %s', mountpoint)
            return

        umnt_p = subprocess.Popen(["umount", mountpoint],
                                  stdout = subprocess.PIPE,
                                  stderr = subprocess.PIPE)
        (p_out, p_err) = umnt_p.communicate()
        if umnt_p.returncode != 0:
            error = _('Unable to unmount filesystem "%s": %s') % (mountpoint,
                                                                  p_err)
            IPlugin.logger.error(error)
            raise OperationFailedEx(error)

        if destroy_mount_point:
            try:
                os.rmdir(mountpoint)
            except:
                pass

    @staticmethod
    def clone(ss_record, fs_record, file_record, dest_file_name=None):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        _validateFileRecord(fs_record, file_record)

        if file_record['fr_type'] != IFileSystemPlugin.FileType:
            error = _('Cannot clone file "%s": Invalid file type "%s"') % \
                (file_record['file_path'], file_record['fr_type'])
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)

        if not _containsFile(ss_record, fs_record,
                             file_record['file_path']):
            error = _('Source file "%s" is not contained on filesystem "%s"') % (
                file_record['file_path'], fs_record['name'])
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)

        if not os.path.exists(file_record['file_path']):
            error = _('Unable to locate file "%s" to clone') % \
                file_record['file_path']
            IPlugin.logger.error(error)
            raise CloneFailedEx(error)

        if dest_file_name:
            if not _containsFile(ss_record, fs_record, dest_file_name):
                error = _('Destination file "%s" is not contained on filesystem "%s"') % (
                    dest_file_name, fs_record['name'])
                IPlugin.logger.error(error)
                raise InvalidValueEx(error)
            if not os.path.exists(os.path.dirname(dest_file_name)):
                os.makedirs(os.path.dirname(dest_file_name))
        else:
            dest_file_name = "%s.%s" % (file_record['file_path'],
                                        time.strftime("%Y%m%d%H%M%S"))

        _callReflink(fs_record, file_record['file_path'], dest_file_name)
        return _fillFileRecord(fs_record, file_path=dest_file_name)

    @staticmethod
    def createSnap(ss_record, fs_record, file_record, snap_name=None):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        _validateFileRecord(fs_record, file_record)

        if file_record['fr_type'] != IFileSystemPlugin.FileType:
            error = _('Cannot snapshot file "%s": Invalid file type "%s"') % \
                (file_record['file_path'], file_record['fr_type'])
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)

        if not _containsFile(ss_record, fs_record,
                             file_record['file_path']):
            error = _('Source file "%s" is not contained on filesystem "%s"') % (
                file_record['file_path'], fs_record['name'])
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)

        if not os.path.exists(file_record['file_path']):
            error = _('Unable to locate file "%s" to snapshot') % \
                file_record['file_path']
            IPlugin.logger.error(error)
            raise CreateSnapFailedEx(error)

        if not snap_name:
            snap_name = time.strftime("%Y%m%d%H%M%S")
        _doSnap(fs_record, file_record['file_path'], snap_name)
        return _fillFileRecord(fs_record,
                               file_path=file_record['file_path'],
                               snap_name=snap_name)

    @staticmethod
    def splitClone(ss_record, fs_record, file_record):
        error = _('Clone split not supported')
        IPlugin.logger.error(error)
        raise NoSuchOperationEx(error)

    @staticmethod
    def cloneFromSnap(ss_record, fs_record, snap_file_record,
                      dest_file_name=None):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        _validateFileRecord(fs_record, snap_file_record)
        if snap_file_record['fr_type'] != IFileSystemPlugin.SnapType:
            error = _('Cannot clone snapshot "%s": Invalid file type "%s"') % \
                (snap_file_record['snap_name'],
                 snap_file_record['fr_type'])
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)

        source_path  = snap_file_record['file_path']
        snap_path = _snapPath(source_path, snap_file_record['snap_name'])
        if not os.path.exists(snap_path) or not _containsFile(ss_record,
                                                              fs_record,
                                                              snap_path):
            error = _('No such snapshot: "%s"' % snap_path)
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)

        if dest_file_name:
            if not _containsFile(ss_record, fs_record, dest_file_name):
                error = _('Destination file "%s" is not contained on filesystem "%s"') % \
                    (dest_file_name, fs_record['name'])
                IPlugin.logger.error(error)
                raise InvalidValueEx(error)
            if not os.path.exists(os.path.dirname(dest_file_name)):
                os.makedirs(os.path.dirname(dest_file_name))
        else:
            dest_file_name = "%s.%s" % (source_path,
                                        time.strftime("%Y%m%d%H%M%S"))

        _callReflink(fs_record, snap_path, dest_file_name)
        return _fillFileRecord(fs_record, file_path=dest_file_name)

    @staticmethod
    def getFileSystemSnapLimits(ss_record, fs_record = None):
        return -1

    @staticmethod
    def getSnapLimits(ss_record, fs_record, file_record):
        if file_record['fr_type'] == IFileSystemPlugin.FileType:
            return -1
        else:
            return 0

    @staticmethod
    def isSnapable(ss_record, fs_record, file_record):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        _validateFileRecord(fs_record, file_record)
        if file_record['fr_type'] != IFileSystemPlugin.FileType:
            return False
        if _containsFile(ss_record, fs_record, file_record["file_path"]):
            return True
        return False

    @staticmethod
    def getFileSystemCloneLimits(ss_record, fs_record = None):
        return -1

    @staticmethod
    def getCloneLimits(ss_record, fs_record, file_record):
        if file_record['fr_type'] == IFileSystemPlugin.FileType:
            return -1
        else:
            return 0

    @staticmethod
    def isCloneable(ss_record, fs_record, file_record):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        _validateFileRecord(fs_record, file_record)
        if file_record['fr_type'] != IFileSystemPlugin.FileType:
            return False
        if _containsFile(ss_record, fs_record, file_record["file_path"]):
            return True

        return False

    @staticmethod
    def isSplittable(ss_record, fs_record, file_record):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        _validateFileRecord(fs_record, file_record)
        if file_record['fr_type'] != IFileSystemPlugin.FileType:
            return False
        if not _containsFile(ss_record, fs_record, file_record["file_path"]):
            return False
        # ocfs2 can't do this anyway ;-)
        return False

    @staticmethod
    def isRestorable(ss_record, fs_record, file_record, snap_file_record):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        _validateFileRecord(fs_record, snap_file_record)

        if file_record['fr_type'] != IFileSystemPlugin.FileType:
            return False
        if snap_file_record['fr_type'] != IFileSystemPlugin.SnapType:
            return False
        if not os.path.samefile(file_record['file_path'],
                                snap_file_record['file_path']):
            return False

        source_path = snap_file_record['file_path']
        snap_path = _snapPath(source_path, snap_file_record['snap_name'])
        return os.path.exists(snap_path) and _containsFile(ss_record,
                                                           fs_record,
                                                           snap_path)

    @staticmethod
    def snapRestore(ss_record, fs_record, snap_file_record):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        _validateFileRecord(fs_record, snap_file_record)
        if snap_file_record['fr_type'] != IFileSystemPlugin.SnapType:
            error = _('Cannot remove restore snapshot "%s": Invalid file type "%s"') % \
                (snap_file_record['snap_name'],
                 snap_file_record['fr_type'])
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)
        source_path = snap_file_record['file_path']
        snap_path = _snapPath(source_path, snap_file_record['snap_name'])
        if not os.path.exists(snap_path) or not _containsFile(ss_record,
                                                              fs_record,
                                                              snap_path):
            error = _('No such snapshot: "%s"' % snap_path)
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)

        if os.path.exists(source_path):
            try:
                os.remove(source_path)
            except OSError, e:
                error = _('Unable to replace restore destination "%s": %s') % (
                    source_path, str(e))
                IPlugin.logger.error(error)
                raise OperationFailedEx(error)

        _callReflink(fs_record, snap_path, source_path)
        return _fillFileRecord(fs_record, file_path=source_path)

    @staticmethod
    def snapRemove(ss_record, fs_record, snap_file_record):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        _validateFileRecord(fs_record, snap_file_record)
        if snap_file_record['fr_type'] != IFileSystemPlugin.SnapType:
            error = _('Cannot remove snapshot "%s": Invalid file type "%s"') % \
                (snap_file_record['snap_name'],
                 snap_file_record['fr_type'])
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)
        snap_path = _snapPath(snap_file_record['file_path'],
                              snap_file_record['snap_name'])
        if not os.path.exists(snap_path) or not _containsFile(ss_record,
                                                              fs_record,
                                                              snap_path):
            error = _('No such snapshot: "%s"' % snap_path)
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)
        else:
            os.remove(snap_path)

    @staticmethod
    def getCurrentClones(ss_record, fs_record, file_record=None):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        if file_record:
            _validateFileRecord(fs_record, file_record)
        raise NoSuchOperationEx('ocfs2 cannot track clones')

    @staticmethod
    def getCurrentSnaps(ss_record, fs_record, file_record=None):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        current_snaps_list = []

        if not file_record:
            error = _('Full scan for snapshots is not implemented')
            IPlugin.logger.error(error)
            raise NotImplementedError(error)
        else:
            _validateFileRecord(fs_record, file_record)

        source_path = file_record['file_path']
        if not os.path.exists(source_path):
            error = _('Cannot determine snapshots for file "%s": '
                      'File does not exist"') % source_path
            IPlugin.logger.error(error)
            raise ListSnapFailedEx(error)

        snap_file_prefix = _snapFileName(os.path.basename(source_path), '')
        source_dir = os.path.dirname(source_path)
        for sfile in os.listdir(source_dir):
            if sfile.startswith(snap_file_prefix):
                file_record = \
                    _fillFileRecord(fs_record, file_path=source_path,
                                    snap_name=_snapNameFromFile(source_path,
                                                                sfile))
                current_snaps_list.append(file_record)

        return current_snaps_list

    @staticmethod
    def create(ss_record, fs_record, file_name, file_type, size, sparse=True, force=False):
        _validateFSRecord(ss_record, fs_record)
        if not fs_record['state'] == FS_STATES.MOUNTED:
            error = _('File system "%s" not mounted') % fs_record['name']
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)

        if file_type == IFileSystemPlugin.FileType:
            create_op = os.mknod
        elif file_type == IFileSystemPlugin.DirType:
            create_op = os.mkdir
        else:
            error = _('Cannot create file of type "%s"') % file_type
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)

        if not file_name.startswith('/'):
            mounts = OCFS2Plugin.listMountPoints(ss_record, fs_record)
            if not mounts:
                error = _('File system "%s" not mounted') % fs_record['name']
                IPlugin.logger.error(error)
                raise InvalidValueEx(error)
            mount = mounts[0]
            file_path = os.path.join(mount['mount_point'], file_name)
        else:
            file_path = file_name
        file_record = _fillFileRecord(fs_record, file_path=file_path, file_sz=0)
        fr = _doCreate(create_op, fs_record, file_record, force)
        if size:
            fr = OCFS2Plugin.resize(ss_record, fs_record, fr, size, sparse)
        return fr

    @staticmethod
    def destroy(ss_record, fs_record, file_record):
        _validateFSRecord(ss_record, fs_record)
        _validateFileRecord(fs_record, file_record)
        _validateFileRecord(fs_record, file_record, file_path=True,
                            fr_type=True)
        if file_record['fr_type'] == IFileSystemPlugin.SnapType:
            error = _('Cannot destroy snapshots')
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)
        if file_record['fr_type'] == IFileSystemPlugin.DirType:
            os.rmdir(file_record['file_path'])
        else:
            os.remove(file_record['file_path'])

    @staticmethod
    def resize(ss_record, fs_record, file_record, new_size, sparse=True):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        _validateFileRecord(fs_record, file_record)
        _validateFileRecord(fs_record, file_record, file_path=True,
                            fr_type=True)
        if file_record['fr_type'] != IFileSystemPlugin.FileType:
            error = _('Cannot resize file of type "%s"') % file_record['fr_type']
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)

        _resizeByVFS(fs_record, file_record, new_size, sparse)
        return OCFS2Plugin.getFileInfo(ss_record, fs_record, file_record)

    @staticmethod
    def getFileInfo(ss_record, fs_record, file_record, file_comp_record=None):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        _validateFileRecord(fs_record, file_record, file_path=True)
        if file_comp_record:
            error = _('Shared extent comparison not yet supported')
            IPlugin.logger.error(error)
            raise NoSuchOperationEx(error)
        if not _containsFile(ss_record, fs_record, file_record['file_path']):
            error = _('File "%s" is not contained on filesystem "%s"') % (
                file_record['file_path'], fs_record['name'])
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)
        st = os.lstat(file_record['file_path'])
        file_record['fr_type'] = _statToFileType(st)
        file_record['ondisk_sz'] = st.st_blocks * 512
        file_record['file_sz' ] = st.st_size
        return _fillFileRecord(fs_record, **file_record)

    @staticmethod
    def list(ss_record, fs_record, file_record, recursive=False):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        _validateFileRecord(fs_record, file_record)
        if file_record['fr_type'] != IFileSystemPlugin.DirType:
            error = _('Not a directory: "%s"') % file_record['file_path']
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)
        list_path = file_record['file_path']
        files = []
        def walk_get_fr(dirname, fname):
            full_path = os.path.join(dirname, fname)
            st = os.lstat(full_path)
            ft = _statToFileType(st)
            oz = st.st_blocks * 512
            fz = st.st_size
            return _fillFileRecord(fs_record, file_path=full_path,
                                   fr_type=ft, ondisk_sz=oz, file_sz=fz)
        if recursive:
            def walk_func(l, dirname, fnames):
                for name in fnames:
                    if file_record.has_key('name_pattern'):
                        if not fnmatch.fnmatch(name, file_record['name_pattern']):
                            continue
                    fr = walk_get_fr(dirname, name)
                    l.append(fr)
        else:
            def walk_func(l, dirname, fnames):
                while len(fnames):
                    name = fnames.pop(0)
                    if file_record.has_key('name_pattern'):
                        if not fnmatch.fnmatch(name, file_record['name_pattern']):
                            continue
                    fr = walk_get_fr(dirname, name)
                    l.append(fr)
        os.path.walk(list_path, walk_func, files)
        return files

    @staticmethod
    def createFileSystem(ss_record, name, backing_device, size=0,
                         access_grp_names=None,
                         qos=None, force=False):
        _validateSSRecord(ss_record)
        if not name:
            error = _('A name is required to create a filesystem')
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)
        if not backing_device:
            error = _('backing_device is required to create a filesystem')
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)
        if backing_device.startswith('/dev/mapper/'):
            mp = backing_device.replace('/dev/mapper/','')
        else:
            if deviceTypeNVMe(backing_device) or deviceTypeMD(backing_device):
                mp = backing_device.replace('/dev/disk/by-id/','')
            else:
                error = _('backing_device type is not supported')
                IPlugin.logger.error(error)
                raise InvalidValueEx(error)

        mp = mp.replace('/','')

        # Multipath devices partition is not allowed, except NVMe and MD Raid devices
        if not deviceTypeNVMe(backing_device) and not deviceTypeMD(backing_device):
            for dev_name in os.listdir(IPlugin.dev_path_prefix):
                if dev_name.startswith(mp + 'p'):
                    error = _('The backing device %s is not allowed to contain partitions') % backing_device
                    IPlugin.logger.error(error)
                    raise InvalidValueEx(error)

        disktype = getPhysicalDiskType(backing_device)

        is_linear_dm = False
        cmd = ['dmsetup', 'ls', '--target', 'linear']
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE, close_fds=True)
        stdoutdata, stderrdata = proc.communicate()
        if proc.returncode == 0:
            for entry in stdoutdata.splitlines():
                devname = entry.split()[0]
                if mp == devname:
                    is_linear_dm = True
                    break
        else:
            IPlugin.logger.error('Error geting linear device mapper name: %s: %s %s',
                                 proc.returncode, stderrdata, stdoutdata)

        fs_label = 'OVS%s' % (name.replace('-', '')[-13:])
        cmd = ['/sbin/mkfs.ocfs2', '-J', 'block64', '-b', str(OCFS2Plugin.FS_BLOCKSIZE),
               '-L', fs_label, '-U', name, '-T', 'vmstore',
               '-N', str(OCFS2Plugin.FS_NODESLOTS)]

        # 'LOCAL' disk type found by getPhysicalDiskType means use mkfs
        #  with 'local' settings, anything else is clustered
        if (disktype and disktype == 'LOCAL') or is_linear_dm:
            cmd.extend(['-M', 'local'])
        if qos:
            cmd.extend(['-T', qos])
        cmd.extend([backing_device])
        if size:
            cmd.extend(['%lu' % _bytesToBlocks(size)])
        mkfs_p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT)
        (p_out, p_none) = mkfs_p.communicate()
        if mkfs_p.returncode != 0:
            if (p_out.find('Overwriting') != -1):
                if force:
                    raise FileSystemBusyEx("Waiting for ocfs2 locking")
                else:
                    error = _('An ocfs2 filesystem already exists on %s') \
                        % backing_device
                    IPlugin.logger.error(error)
                    raise FileSystemBusyEx(error)
            else:
                error = _('Unable to create ocfs2 filesystem on "%s": %s') % \
                    (backing_device, p_out)
                IPlugin.logger.error(error)
                raise OperationFailedEx(error)
        info = _getOneOCFS2Device(backing_device)
        return _deviceToFSRecord(ss_record, info)

    @staticmethod
    def resizeFileSystem(ss_record, fs_record, new_size):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)

        cmd = ['/sbin/tunefs.ocfs2', '-Sbytes', fs_record['backing_device'],
               '%ld' % new_size]
        tunefs_p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT)
        (p_out, p_none) = tunefs_p.communicate()
        if tunefs_p.returncode != 0:
            error = _('Unable to resize ocfs2 filesystem on "%s": %s') % \
                (fs_record['backing_device'], p_out)
            IPlugin.logger.error(error)
            raise OperationFailedEx(error)
        os.system('sync')
        return _fillFSInfo(ss_record, fs_record)

    @staticmethod
    def destroyFileSystem(ss_record, fs_record):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        mounted_error = _('Cannot destroy filesystem "%s": still mounted') % \
                fs_record['name']
        # Quick check
        if fs_record['state'] == FS_STATES.MOUNTED:
            IPlugin.logger.error(mounted_error)
            raise FileSystemBusyEx(mounted_error)
        # Slow check
        if _getFSStatus(ss_record, fs_record) == 'mounted':
            IPlugin.logger.error(mounted_error)
            raise FileSystemBusyEx(mounted_error)
        # Is it even an ocfs2 filesystem?
        found = False
        target_dev = fs_record['backing_device']
        target_dev = target_dev[target_dev.rfind('/') + 1:]
        for dev in _getOCFS2Devices():
            if dev.endswith(target_dev):
                found = True
        if not found:
            error = _('Cannot destroy filesystem "%s" on device "%s": No filesystem on device') % (
                fs_record['name'], fs_record['backing_device']
            )
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)
        fd = os.open(fs_record['backing_device'], os.O_WRONLY)
        # Wipe the superblock.  Cover the maximum ocfs2 block size
        os.write(fd, '\0' * 4096 * 3)
        os.fsync(fd)
        os.close(fd)

    @staticmethod
    def startPresent(ss_record, fs_record, access_grp_names):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        if len(access_grp_names) != 1:
            error = 'Invalid access group names: "%s"' % access_grp_names
            raise InvalidValueEx(error)

        cmd = ['/sbin/o2cluster', '-y',
               '--update=o2cb,%s,global' % access_grp_names[0],
               fs_record['backing_device']]

        o2cluster_p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                       stderr=subprocess.STDOUT)
        (p_out, p_none) = o2cluster_p.communicate()
        if o2cluster_p.returncode != 0:
            error = _('Unable to update ocfs2 cluster on "%s": %s') % \
                (fs_record['backing_device'], p_out)
            IPlugin.logger.error(error)
            raise OperationFailedEx(error)
        return _fillFSInfo(ss_record, fs_record)

    @staticmethod
    def stopPresent(ss_record, fs_record, access_grp_names):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)

        cmd = ['/sbin/o2cluster', '-y', '--update=o2cb,NOTINUSE,global',
               fs_record['backing_device']]

        o2cluster_p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                       stderr=subprocess.STDOUT)
        (p_out, p_none) = o2cluster_p.communicate()
        if o2cluster_p.returncode != 0:
            error = _('Unable to update ocfs2 cluster on "%s": %s') % \
                (fs_record['backing_device'], p_out)
            IPlugin.logger.error(error)
            raise OperationFailedEx(error)
        return _fillFSInfo(ss_record, fs_record)

#
# File local helpers.  These should never appear in __all__.
#

def _validateRecord(name, record, key_hash):
    '''Validate a record.  Any key in key_hash must exist in the record.
       if key_hash[key] is True, the record must have a non-False value
       for it.'''
    for key in key_hash:
        if not record.has_key(key):
            raise MissingKeyEx(_('%s does not have key "%s"') % (name, key),
                               key)
        elif key_hash[key] and not record[key]:
            raise MissingKeyEx(_('%s does not have a value for key "%s"') %
                               (name, key), key)

def _buildKeyHash(check_keys, require_value_keys, record_fields):
    '''Build a key_hash for _validateRecord().  record_fields is only used
       if there are no keys in the other lists.  It is the record
       definition, and if record_fields[key][1] is true then key is treated
       like it exists in check_keys.'''
    key_hash = {}
    for key in check_keys:
        key_hash[key] = False
    for key in require_value_keys.iterkeys():
        key_hash[key] = require_value_keys[key]
    if not len(key_hash.keys()):
        for (key, value) in record_fields.iteritems():
            if value[1]:
                key_hash[key] = False
    return key_hash

def _validateSSRecord(ss_record, *check_keys, **require_value_keys):
    key_hash = _buildKeyHash(check_keys, require_value_keys,
                             IFileSystemPlugin.__SSRecord_fields__)
    _validateRecord('SSRecord', ss_record, key_hash)
    if ss_record["storage_type"] != IFileSystemPlugin.FileSysStorage:
        raise InvalidValueEx(_('Invalid storage type: "%s"') %
                             ss_record['storage_type'])

def _validateFSRecord(ss_record, fs_record, *check_keys,
                      **require_value_keys):
    key_hash = _buildKeyHash(check_keys, require_value_keys,
                             IFileSystemPlugin.__FSRecord_fields__)
    _validateRecord('FSRecord', fs_record, key_hash)
    if key_hash.has_key('ss_uuid') and \
        fs_record['ss_uuid'] != ss_record['uuid']:
        raise InvalidValueEx(_('Invalid Storage Server uuid: expected "%s", got "%s"') %
                             (ss_record['uuid'], fs_record['ss_uuid']))

def _validateMountRecord(fs_record, mount_record, *check_keys,
                         **require_value_keys):
    key_hash = _buildKeyHash(check_keys, require_value_keys,
                             IFileSystemPlugin.__MountRecord_fields__)
    _validateRecord('MountRecord', mount_record, key_hash)
    if fs_record and key_hash.has_key('fs_uuid') and \
        mount_record['fs_uuid'] != fs_record['uuid']:
        raise InvalidValueEx(_('Invalid Filesystem uuid: expected "%s", got "%s"') %
                             (fs_record['uuid'], mount_record['ss_uuid']))

def _validateFileRecord(fs_record, file_record, *check_keys,
                        **require_value_keys):
    key_hash = _buildKeyHash(check_keys, require_value_keys,
                             IFileSystemPlugin.__FileRecord_fields__)
    _validateRecord('FileRecord', file_record, key_hash)
    if key_hash.has_key('fs_uuid') and \
        file_record['fs_uuid'] != fs_record['uuid']:
        raise InvalidValueEx(_('Invalid Filesystem uuid: expected "%s", got "%s"') %
                             (fs_record['uuid'], file_record['ss_uuid']))
    if key_hash.has_key('file_path') and \
        file_record['file_path'][0] != '/':
        raise InvalidValueEx(_('Path "%s" is not absolute') %
                             file_record['file_path'])


def _statToFileType(st):
    if stat.S_ISDIR(st.st_mode):
        return IFileSystemPlugin.DirType
    elif stat.S_ISBLK(st.st_mode):
        return IFileSystemPlugin.BlkDevType
    elif stat.S_ISCHR(st.st_mode):
        return IFileSystemPlugin.CharDevType
    elif stat.S_ISLNK(st.st_mode):
        return IFileSystemPlugin.SymLinkType
    elif stat.S_ISSOCK(st.st_mode):
        return IFileSystemPlugin.SockFileType
    elif stat.S_ISFIFO(st.st_mode):
        return IFileSystemPlugin.FifoType
    else:
        return IFileSystemPlugin.FileType

# Yes, I know I should write a generator.
def _procMountsToMount(line):
    l = line.split()
    return {
        'device': l[0].decode('string_escape'),
        'mountpoint': l[1].decode('string_escape'),
        'fstype': l[2],
        'options': l[3].decode('string_escape')
    }

def _deviceToFSRecord(ss_record, device):
    fr = _fillFSRecord(ss_record, backing_device=device['device'],
                       name=device.get('label', ''), uuid=device['uuid'],
                       size=device['size'], free_sz=device['free'], cluster_name=device['cluster_name'])
    return fr

def _getMountUUID(fs_uuid, mountpoint):
    uuid_url = 'scmount:%s:%s' % (fs_uuid, mountpoint)
    return str(uuid.uuid5(uuid.NAMESPACE_URL, uuid_url))

def _deviceAndMountToMountRecord(device, mount):
    uuid_url = 'scmount:%s:%s' % (device['uuid'], mount['mountpoint'])
    return _fillMountRecord(None, uuid=_getMountUUID(device['uuid'],
                                                     mount['mountpoint']),
                            fs_uuid=device['uuid'],
                            mount_point=mount['mountpoint'],
                            options=mount['options'].split(','),
                            status='mounted')

def _getMounts():
    mounts = {}
    try:
        try:
            proc_mounts = open("/proc/mounts", "r")
            for line in proc_mounts:
                mount = _procMountsToMount(line)
                if mount['fstype'] != 'ocfs2':
                    continue
                if mounts.has_key(mount['device']):
                    mounts[mount['device']].append(mount)
                else:
                    mounts[mount['device']] = [mount]
        except IOError, ioe:
            proc_mounts.close()
            raise
    finally:
        proc_mounts.close()
    return mounts

def _getOneOCFS2Device(device):
    uuid = None
    label = None
    clustersize = None
    size = None
    free = None
    cluster_name = None

    debugfs_p = subprocess.Popen(['/sbin/debugfs.ocfs2', '-n', device],
                                 stdin = subprocess.PIPE,
                                 stdout = subprocess.PIPE,
                                 stderr = subprocess.PIPE)
    (p_out, p_err) = debugfs_p.communicate('stats\nstat -T //global_bitmap\nquit')
    if debugfs_p.returncode != 0:
        error = _('Unable to query ocfs2 device "%s"') % device
        IPlugin.logger.error('_getOneOCFS2Device() - %s' % error)
        raise OperationFailedEx(error)
    if p_err:
        error = _('Unable to query ocfs2 device "%s": %s' % (device,
                                                             p_err))
        IPlugin.logger.error('_getOneOCFS2Device() - %s' % error)
        raise OperationFailedEx(error)
    r = StringIO(p_out)
    try:
        for line in r:
            if line.find(':') == -1:
                continue
            if not clustersize:
                if line.find('Cluster Size Bits:') != -1:
                    bits = line.split(':')[2].strip()
                    clustersize = pow(2, int(bits))
                    continue
            if not uuid:
                if line.strip().split(':')[0] == 'UUID':
                    uuid = line.split(':')[1].strip().lower()
                    continue
            if not label:
                if line.split(':')[0].strip() == 'Label':
                    label = line.split(':')[1][1:].rstrip('\n')
                    continue
            if not cluster_name:
                if line.split(':')[0].strip() == 'Cluster name':
                    cluster_name = line.split(':')[1][1:].rstrip('\n')
                    continue
            if not size:
                if line.split(':')[0].strip() == 'Bitmap Total':
                    clusters = line.split(':')[1].split()[0]
                    size = long(clusters) * clustersize

		    l = line.find('Free:')
                    freecs = line[l+5:].rstrip()
                    free = long(freecs) * clustersize
                    continue
    finally:
        r.close()

    if cluster_name == None:
        cluster_name = 'NONCLUSTERED'

    if uuid != None and label != None and size != None:
        return {'device': device, 'uuid': uuid, 'label': label,
                'size': size, 'free': free, 'cluster_name': cluster_name}
    else:
        error = _('Problem querying ocfs2 devices "%s"') % device
        IPlugin.logger.error(error)
        raise OperationFailedEx(error)

def _getOCFS2Devices():
    devs = []
    if not os.path.exists(IPlugin.dev_path_prefix):
        return devs

    # Only get devices with active paths (keep cmds like /sbin/blkid from failing)
    if os.path.exists("/dev/emcpower"):
        dev_names = getActiveDevPP()
    else:
        dev_names = getActiveDevMapper()

    sys_repo = glob.glob(os.path.join(IPlugin.dev_path_prefix, "OVM_SYS_REPO*"))
    if sys_repo:
       dev_names.append(sys_repo[0])

    nvme_devs = glob.glob(os.path.join(IPlugin.dev_path_by_id, "nvme-*"))
    if nvme_devs:
        dev_names = dev_names + nvme_devs
    md_devs = glob.glob(os.path.join(IPlugin.dev_path_by_id, "md-uuid-*"))
    if md_devs:
        dev_names = dev_names + md_devs

    if not dev_names:
        return devs

    for device in dev_names:

        blkid_c =["/sbin/blkid",
              "-t", "TYPE=ocfs2",
              "-o", "device", "%s" % device] 

        blkid_p = subprocess.Popen(blkid_c,
                               stdout = subprocess.PIPE,
                               stderr = subprocess.PIPE)
        (p_out, p_err) = blkid_p.communicate()

        if blkid_p.returncode == 0:
            r = StringIO(p_out)
            try:
                for dev in r:
                    devs.append(dev.rstrip('\n'))
            finally:
                r.close()
        elif blkid_p.returncode != 2:
            IPlugin.logger.error("_getOCFS2Devices() - Unable to query " +
                             "ocfs2 devices")
            raise OperationFailedEx(_("Unable to query ocfs2 devices"))

    return devs

def _getFSStatus(ss_record, fs_record):
    _validateSSRecord(ss_record)
    _validateFSRecord(ss_record, fs_record)
    if len(OCFS2Plugin.listMountPoints(ss_record, fs_record)):
        fs_record['status'] = 'mounted'
        fs_record['state'] = FS_STATES.MOUNTED

    else:
        fs_record['status'] = 'not mounted'
        fs_record['state'] = FS_STATES.UNMOUNTED
    return fs_record['status']

def _fillFSInfo(ss_record, fs_record):
    _validateSSRecord(ss_record)
    _validateFSRecord(ss_record, fs_record, "backing_device", "ss_uuid")
    info = _getOneOCFS2Device(fs_record['backing_device'])
    if info['label']:
        fs_record['name'] = info['label']
    else:
        fs_record['name'] = info['uuid']
    fs_record['cluster_name'] = info['cluster_name']
    fs_record['uuid'] = info['uuid']
    fs_record['size'] = info['size']
    fs_record['free_sz'] = info['free']
    fs_record['status'] = 'unknown'
    fs_record['state'] = FS_STATES.UNKNOWN
    fs_record["access_grp_names"] = "all"
    mounts = OCFS2Plugin.listMountPoints(ss_record, fs_record)
    if len(mounts):
        fs_record['state'] = FS_STATES.MOUNTED
        fs_record['status'] = 'mounted'
        fs_record['mount_options'] = mounts[0]['options']
    else:
        fs_record['state'] = FS_STATES.UNMOUNTED
        fs_record['status'] = 'not mounted'
    return fs_record

def _getModuleStatus():
    f = file("/proc/filesystems", "r")
    try:
        for l in f:
            if l.split()[0] == 'ocfs2':
                return "loaded"
        return "not loaded"
    finally:
        f.close()

def _fillSSInfo(ss_record):
    ss_record['total_sz'] = -1
    ss_record['used_sz'] = -1
    ss_record['free_sz'] = -1
    ss_record['access_grps'] = []
    ss_record['storage_desc'] = "The ocfs2 Filesystem"
    ss_record['status'] = _getModuleStatus()
    return ss_record

def _fillSSRecord(**kwargs):
    '''Take attributes for an ss_record and fill it in.  This makes sure
       everything needed was specified.'''
    ss_record = kwargs
    if not ss_record.has_key('access_grps'):
        ss_record['access_grps'] = []
    if not ss_record.has_key('storage_desc'):
        ss_record['storage_desc'] = 'The ocfs2 filesystem'
    for key in ['admin_host', 'admin_user', 'admin_passwd', 'total_sz',
                'used_sz', 'free_sz', 'status']:
        if not ss_record.has_key(key):
            ss_record[key] = IFileSystemPlugin.__SSRecord__[key]
    _validateSSRecord(ss_record)
    return ss_record

def _fillFSRecord(ss_record, **kwargs):
    '''Take attributes for a fs_record and fill it in.  This makes sure
       everything needed was specified.'''
    _validateSSRecord(ss_record)
    fs_record = kwargs
    fs_record['ss_uuid'] = ss_record['uuid']
    fs_record['access_grp_names'] = 'all'
    if not fs_record.get('name', ''):
        fs_record['name'] = fs_record['uuid']
    if not fs_record.has_key('status'):
        fs_record['status'] = 'unknown'
    if not fs_record.has_key('state'):
        fs_record['state'] = FS_STATES.UNKNOWN
    for key in ['size']:
        if not fs_record.has_key(key):
            fs_record[key] = IFileSystemPlugin.__FSRecord__[key]
    _validateFSRecord(ss_record, fs_record)
    return fs_record

def _fillMountRecord(fs_record, **kwargs):
    '''Take attributes for a mount record and fill it in.  This makes sure
       everything needed was specified'''
    mount_record = kwargs
    if mount_record.get('status', ''):
        mount_record['status'] = 'unknown'
    _validateMountRecord(fs_record, mount_record)
    return mount_record

def _fillFileRecord(fs_record, **kwargs):
    '''Take attributes for a file record and fill it in.  This makes sure
       everything needed was specified'''
    file_record = kwargs
    file_record['fs_uuid'] = fs_record['uuid']
    if file_record.has_key('snap_name'):
        file_record['fr_type'] = IFileSystemPlugin.SnapType
    elif not file_record.has_key('fr_type'):
        file_record['fr_type'] = IFileSystemPlugin.FileType
    for key in ['ondisk_sz', 'file_sz']:
        if not file_record.has_key(key):
            file_record[key] = IFileSystemPlugin.__FileRecord__[key]
    _validateFileRecord(fs_record, file_record)
    return file_record

def _callReflink(fs_record, src_path, dest_path):
    refln_p = subprocess.Popen(["reflink", '-P', src_path, dest_path],
                               stdout = subprocess.PIPE,
                               stderr = subprocess.PIPE)
    (p_out, p_err) = refln_p.communicate()

    if refln_p.returncode != 0:
        error = _('Unable to clone file "%s" to "%s": %s') % (
            src_path, dest_path, p_err)
        IPlugin.logger.error(error)
        raise CloneFailedEx(error)

# XXX This will use snap.ocfs2 in the future, these are not
#     production locations.
def _snapFileName(src_file, snap_name):
    return "%s.snap.%s" % (src_file, snap_name)

def _snapNameFromFile(src_file, snap_file):
    snap_file_prefix = _snapFileName(os.path.basename(src_file), '')
    return snap_file[len(snap_file_prefix):]

def _snapPath(src_path, snap_name):
    if not snap_name or '/' in snap_name:
        error = _('Invalid snapshot name: "%s"') % snap_name
        IPlugin.logger.error(error)
        raise InvalidValueEx(error)
    return _snapFileName(src_path, snap_name)

def _doSnap(fs_record, src_path, snap_name):
    dest_path = _snapPath(src_path, snap_name)
    try:
        _callReflink(fs_record, src_path, dest_path)
        os.chmod(dest_path, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)
    except CloneFailedEx, e:
        error = _('Unable to create snapshot "%s" for file "%s"') % (
            snap_name, src_path)
        IPlugin.logger.error(error)
        raise CreateSnapFailedEx(error)
    except OSError, e:
        error = _('Unable to set permissions on snapshot "%s" for file "%s"') % (
            snap_name, src_path)
        IPlugin.logger.error(error)
        raise CreateSnapFailedEx(error)

def _doCreate(create_op, fs_record, file_record, force):
    try:
        create_op(file_record['file_path'])
        return _fillFileRecord(fs_record, **file_record)
    except OSError, e:
        if e.errno != errno.EEXIST or not force:
            raise
    if stat.S_ISDIR(os.stat(file_record['file_path']).st_mode):
            os.rmdir(file_record['file_path'])
    else:
            os.remove(file_record['file_path'])
    create_op(file_record['file_path'])
    return _fillFileRecord(fs_record, **file_record)

def _callRESVioctl(fd, old_size, new_size):
    arch = platform.architecture()[0]
    if arch == '64bit':
        ioc = 0x4030582a
    else:
        ioc = 0x402c582a
    # Note that we can't handle 32bit-on-64bit until ocfs2 does.
    arg = struct.pack('@4x2q24x', old_size, new_size - old_size)
    ioctl(fd, ioc, arg)

def _growNonSparse(fd, old_size, new_size):
    try:
        _callRESVioctl(fd, old_size, new_size)
        os.ftruncate(fd, new_size)
        return
    except OSError, e:
        if e.errno != errno.ENOTTY:
            raise

    os.lseek(fd, 0, os.SEEK_END)
    left = new_size - old_size
    while left > 0:
        count = pow(2, 20)
        if count > left:
            count = left
        os.write(fd, '\0' * count)
        left -= count
    os.fsync(fd)

def _resizeByVFS(fs_record, file_record, new_size, sparse):
    if int(file_record['file_sz']) == new_size:
        return

    fd = os.open(file_record['file_path'], os.O_WRONLY)
    try:
        if (new_size < int(file_record['file_sz'])) or sparse:
            os.ftruncate(fd, new_size)
        else:
            _growNonSparse(fd, int(file_record['file_sz']), new_size)
    finally:
        os.close(fd)

def _bytesToBlocks(in_bytes):
    return ((in_bytes + (OCFS2Plugin.FS_BLOCKSIZE - 1)) /
            OCFS2Plugin.FS_BLOCKSIZE)

def _containsFile(ss_record, fs_record, file_path):
    """
    Check if the file is contained on the file system identified by the
    File System record.

    @param fs_record: L{File System record<OSCPlugin.IPlugin.__FSRecord__>}
    @type  fs_record: C{dict}

    @param file_path: File path to check.
    @type  file_path: C{str}

    @return: True if the file is contained in the file system.
    @rtype:  C{bool}

    @raise IPluginException: on failure
    """
    if not file_path.startswith('/'):
        raise InvalidValueEx(_('Path "%s" is not absolute') % file_path)
    mounts = OCFS2Plugin.listMountPoints(ss_record, fs_record)
    if not len(mounts):
        raise OperationPreReqNotMetEx(_('Filesystem "%s" is not mounted') %
                                      fs_record['name'])
    mount_stat = os.stat(mounts[0]['mount_point'])
    while True:
        try:
            file_stat = os.stat(file_path)
            return mount_stat.st_dev == file_stat.st_dev
        except OSError, e:
            if e.errno != errno.ENOENT:
                raise
        new_path = os.path.dirname(file_path)
        # Have we just tried '/'?
        if new_path == file_path:
            break
        file_path = new_path
    return False
